import 'package:contentsquare/src/modules/webview_tracking/logs/initialization_error_log.dart';
import 'package:contentsquare/src/modules/webview_tracking/webview_flutter_tracking.dart';
import 'package:contentsquare/src/modules/webview_tracking/screengraph_retrieval.dart';
import 'package:contentsquare/src/modules/webview_tracking/webview_tracking.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:webview_flutter/webview_flutter.dart';

import '../../../helpers/expect_logs.dart';
import '../../../helpers/matchers/fixtures/matcher.dart';
import '../../../helpers/mocks/fake_webview.dart';

class _MockWebViewController extends Mock implements WebViewController {}

void main() {
  ContentsquareFlutterWebViewTracker buildDummyWebViewTrackerImplementation() {
    return ContentsquareFlutterWebViewTracker(
      ContentsquareWebViewTrackerImpl(
        screengraphRetrievalConfiguration: ScreengraphRetrievalConfiguration(
          maxExecutionDuration: const Duration(seconds: 10),
          intervalBetweenJSCalls: const Duration(milliseconds: 100),
        ),
        getEncodedWebViewTrackerTVP: () => null,
      ),
    );
  }

  group(
    '$ContentsquareWebViewTrackerBuilder',
    () {
      test(
        'should log a a $WebViewInitializationErrorLog on any other platform than Android or IOS',
        () async {
          debugDefaultTargetPlatformOverride = TargetPlatform.fuchsia;

          final fakeWebViewController = FakeWebViewController(
            onRunJavascriptReturningResult: (_) {
              return Future.value('');
            },
          );

          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.onWebViewCreated(fakeWebViewController);

          expectLogs(
            (log) => expect(log, isA<WebViewInitializationErrorLog>()),
          );

          await tracker.onPageFinished('url');
        },
      );

      test(
        'should return normally on Android when the methods are called in the right order and the javascript channel are initialized correctly',
        () async {
          debugDefaultTargetPlatformOverride = TargetPlatform.android;
          const _androidJavascriptTrueString = 'true';

          final fakeWebViewController = FakeWebViewController(
            onRunJavascriptReturningResult: (_) {
              return Future.value(_androidJavascriptTrueString);
            },
          );

          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.onWebViewCreated(fakeWebViewController);

          await expectLater(
            () => tracker.onPageFinished('url'),
            returnsNormally,
          );
        },
      );

      test(
        'should return normally on iOS when the methods are called in the right order and the javascript channel are initialized correctly',
        () async {
          debugDefaultTargetPlatformOverride = TargetPlatform.iOS;
          const _iosJavascriptTrueString = '1';

          final fakeWebViewController = FakeWebViewController(
            onRunJavascriptReturningResult: (_) {
              return Future.value(_iosJavascriptTrueString);
            },
          );

          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.onWebViewCreated(fakeWebViewController);

          await expectLater(
            () => tracker.onPageFinished('url'),
            returnsNormally,
          );
        },
      );

      test(
        'The bridge implementation should match its fixture',
        () async {
          debugDefaultTargetPlatformOverride = TargetPlatform.iOS;
          const _iosJavascriptTrueString = '1';

          final mockWebViewController = _MockWebViewController();
          late String bridge;
          when(
            () => mockWebViewController.runJavascriptReturningResult(any()),
          ).thenAnswer((invocation) async {
            bridge = invocation.positionalArguments.first;

            return _iosJavascriptTrueString;
          });
          when(
            () => mockWebViewController.currentUrl(),
          ).thenAnswer((_) async => 'url');

          final tracker = buildDummyWebViewTrackerImplementation();

          tracker.onWebViewCreated(mockWebViewController);
          await tracker.onPageFinished('url');

          await expectLater(
            bridge,
            matchesFixture(
              'expectedBridgeJSToDartToInject',
              isVersionSpecific: false,
            ),
          );
        },
      );
    },
  );
}
